@@ -1,19 +1,22 @@
 import importlib
 import logging
 import os
+import urllib.parse
 from logging import getLogger
 
 import requests
 from django.conf import settings
 from django.contrib.auth import get_user_model
-from django.db import models
-from django.utils import translation
-from rest_framework import authentication
+from django.core.cache import cache
+from django.http.request import HttpRequest
+from django.utils import timezone, translation
 from rest_framework.request import Request
+from rest_framework_simplejwt import authentication
+from rest_framework_simplejwt.exceptions import InvalidToken
 
 from . import custom_exceptions
 from .translator import translate as _
-from .utils import get_language_code
+from .utils import decode_jwt_access_token, get_language_code
 
 logger = getLogger(__file__)
 
@@ -111,21 +114,9 @@ get_user: callable = load_callable(LEARNGUAL_AUTH_GET_USER)
 assert get_user, "No callable exists in path:%s" % LEARNGUAL_AUTH_GET_USER
 
 
-class LearngualAuthentication(authentication.BaseAuthentication):
-    """
-    An authentication plugin that authenticates requests through a JSON web
-    token provided in a request header.
-    """
-
-    www_authenticate_realm = "api"
-    media_type = "application/json"
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self.user_model = get_user_model()
-
+class AuthenticationBaseHelper:
     @classmethod
-    def get_http_headers(cls, request: Request):
+    def get_http_headers(cls, request: Request | HttpRequest):
         headers = {
             key.lower().lstrip("http_"): value
             for key, value in request.META.items()
@@ -134,17 +125,13 @@ class LearngualAuthentication(authentication.BaseAuthentication):
         return headers
 
     @classmethod
-    def get_query_str(cls, request: Request):
+    def get_query_str(cls, request: Request | HttpRequest):
         return "?{query}".format(query=request.META.get("QUERY_STRING", ""))
 
-    def authenticate(self, request):
-        # this is for test mode
-        service_key = str(
-            request.META.get("HTTP_SERVICE_KEY", "")
-            or request.GET.get("service-key", "")
-            or request.GET.get("_service-key", "")
-        )
-        if service_key:
+    def authenticate_service(self, request: Request | HttpRequest):
+        headers = self.get_header(request)
+
+        if service_key := headers.get("service_key"):
             logger.info("authentication send with service account")
             if service_key == LEARNGUAL_SERVICE_API_KEY:
                 service_user = User(id="service", username="service")
@@ -154,21 +141,124 @@ class LearngualAuthentication(authentication.BaseAuthentication):
             msg = _("Invalid service key")
             raise custom_exceptions.UnAuthenticated(msg)
 
-        header = self.get_header(request)
-        logger.info(
-            "retrieve required authentication header during authentication -> %s",
-            header,
+    def get_user(self, validated_token, headers):
+        """
+        Attempts to find and return a user using the given validated token.
+        """
+        try:
+            user_id = validated_token[authentication.api_settings.USER_ID_CLAIM]
+        except KeyError:
+            raise InvalidToken(_("Token contained no recognizable user identification"))
+
+        res_data = {}
+        res_data["id"] = user_id
+        if account_id := headers.get("account_id"):
+            res_data["account"] = {"id": account_id}
+        user = get_user(res_data)
+        account = getattr(user, "account", None)
+        if account and account.owner != user:
+            logger.warning(f"{account = }, {account.owner = }, request user {user}")
+            raise custom_exceptions.UnAuthenticated(_("Invalid account provided"))
+        return user
+
+    def get_raw_token(self, header):
+        """
+        Extracts an unvalidated JSON web token from the given "Authorization"
+        header value.
+        """
+        parts = header.split()
+
+        if len(parts) == 0:
+            # Empty AUTHORIZATION header sent
+            return None
+
+        if str(parts[0]).lower().strip() != "bearer":
+            # Assume the header does not contain a JSON web token
+            return None
+
+        if len(parts) != 2:
+            raise custom_exceptions.UnAuthenticated(
+                _("Authorization header must contain two space-delimited values"),
+                code="bad_authorization_header",
+            )
+
+        return parts[1]
+
+    def append_bearer(self, token: str):
+        if (
+            token
+            and isinstance(token, str)
+            and not token.lower().strip().startswith("bearer")
+        ):
+            return "Bearer " + token
+        return token
+
+    def get_header(self, request: Request | HttpRequest):
+        """
+        Extracts the header containing the JSON web token from the given
+        request.
+        """
+        account_id = str(
+            request.META.get("HTTP_ACCOUNT", "")
+            or request.META.get("HTTP_X_ACCOUNT", "")
+            or request.COOKIES.get("x-account", "")
+            or request.COOKIES.get("account", "")
+            or request.COOKIES.get("_account", "")
+            or request.GET.get("x-account")
+            or request.GET.get("_account")
+            or ""
         )
-        if not header.get("authorization"):
-            logger.warning("no authorization")
+        api_key = str(
+            request.META.get("HTTP_API_KEY", "")
+            or request.META.get("HTTP_X_API_KEY", "")
+            or request.COOKIES.get("x-api-key", "")
+            or request.COOKIES.get("api-key", "")
+            or request.GET.get("x-api-key")
+            or request.GET.get("api-key")
+            or ""
+        )
+        service_key = str(
+            request.META.get("HTTP_SERVICE_KEY", "")
+            or request.META.get("HTTP_X_SERVICE_KEY", "")
+            or request.COOKIES.get("x-service-key", "")
+            or request.COOKIES.get("service-key", "")
+            or request.GET.get("service-key", "")
+            or request.GET.get("_service-key", "")
+            or ""
+        )
+        authorization = str(
+            request.META.get("HTTP_AUTHORIZATION", "")
+            or request.COOKIES.get("x-authorization", "")
+            or request.COOKIES.get("authorization", "")
+            or request.COOKIES.get("x-token", "")
+            or request.COOKIES.get("token", "")
+            or request.GET.get("token", "")
+            or request.GET.get("_token", "")
+            or ""
+        )
+
+        return dict(
+            account_id=urllib.parse.unquote(account_id or ""),
+            api_key=urllib.parse.unquote(api_key or ""),
+            authorization=urllib.parse.unquote(self.append_bearer(authorization or "")),
+            service_key=urllib.parse.unquote(service_key or ""),
+        )
+
+
+class LearngualAuthentication(
+    AuthenticationBaseHelper, authentication.JWTAuthentication
+):
+    """
+    An authentication plugin that authenticates requests through a JSON web
+    token provided in a request header.
+    """
+
+    def authenticate_api_key(self, request: Request | HttpRequest):
+        header = self.get_header(request)
+        if not (api_key := header.get("api_key")):
             return
-
         headers = self.get_http_headers(request)
-        logger.info("get django request header for forwarding -> %s", headers)
-
         query_str = self.get_query_str(request)
-        logger.info("query string for forwarding -> %s", query_str)
-        logger.info(f"{LEARNGUAL_AUTH_RETRIEVE_URL =}")
         res = requests.get(LEARNGUAL_AUTH_RETRIEVE_URL + query_str, headers=headers)
         if not res.ok:
             logging.warning(
@@ -178,131 +268,58 @@ class LearngualAuthentication(authentication.BaseAuthentication):
                 raise custom_exceptions.UnAuthenticated(_("Authentication required"))
             raise custom_exceptions.UnAuthenticated(_("Service is down."))
         res_data = res.json()
-        user = get_user(res_data)
+        if not (user := get_user(res_data)):
+            return
         self.set_language(user, request)
-        return (
-            user,
-            header.get("api_key") or header.get("authorization").split(" ")[-1],
-        )
+        return (user, api_key)
 
-    def set_language(self, user, request):
+    def authenticate(self, request: Request | HttpRequest):
+        if is_service := self.authenticate_service(request):
+            return is_service
+        if is_api := self.authenticate_api_key(request):
+            return is_api
+        header = self.get_header(request)
+        if not (authorization := header.get("authorization")):
+            return
+        raw_token = self.get_raw_token(authorization)
+        if raw_token is None:
+            return
+        try:
+            validated_token = self.get_validated_token(raw_token)
+        except InvalidToken as e:
+            raise custom_exceptions.UnAuthorized(*e.args)
+        data = decode_jwt_access_token(str(validated_token))
+        jti = data.get("jti")
+        anonymous_id = data.get("anonymous_id", None)
+        if jti and cache.get(jti):
+            raise custom_exceptions.UnAuthorized("token blacklisted")
+
+        if not (user := self.get_user(validated_token, header)):
+            return
+
+        if user and user.token_blacklist_at:
+            created_at: str = data.get("created_at")
+            if (
+                not created_at
+                or timezone.datetime.fromisoformat(created_at.rstrip("zZ"))
+                <= user.token_blacklist_at
+            ):
+                raise custom_exceptions.UnAuthorized("token blacklisted")
+        self.set_language(user, request)
+        if (
+            user
+            and anonymous_id
+            and (update_anonymous_link := getattr(user, "update_anonymous_link", None))
+        ):
+            if callable(update_anonymous_link):
+                update_anonymous_link(anonymous_id)
+            else:
+                raise custom_exceptions.UnAuthenticated("Invalid anonymous user")
+        return user, validated_token
+
+    def set_language(self, user, request: Request | HttpRequest):
         if user and not request.GET.get("_lang"):
             if account := getattr(user, "account", None):
                 translation.activate(get_language_code(account.language))
             else:
                 translation.activate(get_language_code(user.language))
-
-    def get_header(self, request):
-        """
-        Extracts the header containing the JSON web token from the given
-        request.
-        """
-        account_id = str(
-            request.META.get("HTTP_ACCOUNT", "")
-            or request.META.get("HTTP_X_ACCOUNT", "")
-        )
-        api_key = str(
-            request.META.get("HTTP_API_KEY", "")
-            or request.META.get("HTTP_X_API_KEY", "")
-        )
-        service_key = str(
-            request.META.get("HTTP_SERVICE_KEY", "")
-            or request.META.get("HTTP_X_SERVICE_KEY", "")
-            or request.GET.get("service-key", "")
-            or request.GET.get("_service-key", "")
-        )
-        authorization = str(request.META.get("HTTP_AUTHORIZATION", ""))
-
-        return dict(
-            account_id=account_id,
-            api_key=api_key,
-            authorization=authorization,
-            service_key=service_key,
-        )
-
-
-def default_user_authentication_rule(user):
-    # Prior to Django 1.10, inactive users could be authenticated with the
-    # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
-    # prevents inactive users from authenticating.  App designers can still
-    # allow inactive users to authenticate by opting for the new
-    # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
-    # users from authenticating to enforce a reasonable policy and provide
-    # sensible backwards compatibility with older Django versions.
-    return user is not None and user.is_active
-
-
-class LearngualFakeAuthentication(authentication.BaseAuthentication):
-    """
-    A fake authentication plugin that authenticates requests through a JSON web
-    token provided in a request header.
-    It simulates authentication by the IAM service, but uses the "account"
-    header to internally retrieve the account
-    """
-
-    www_authenticate_realm = "api"
-    media_type = "application/json"
-    account_model: type[models.Model] | None = None
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self.user_model = get_user_model()
-        self.account_model = load_callable(LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH)
-
-    @classmethod
-    def get_http_headers(cls, request: Request):
-
-        headers = dict()
-        if request:
-            for key, value in request.META.items():
-                if key.lower().startswith("http_"):
-                    headers[key.lower().lstrip("http_")] = value
-        return headers
-
-    @classmethod
-    def get_query_str(cls, request: Request):
-        return "?{query}".format(query=request.META.get("QUERY_STRING", ""))
-
-    def authenticate(self, request):
-        logger.info("authenticate() starting")
-        header = self.get_header(request)
-
-        headers = self.get_http_headers(request)
-
-        account_or_user_id = headers.get("account")
-        # logger.info("account_or_user_id: %s", account_or_user_id)
-        account = None
-        user = None
-        if self.account_model:
-            account = self.account_model.objects.filter(id=account_or_user_id).first()
-            if account:
-                logger.info("account exist: %s", account)
-                user = getattr(account, "owner", None) or getattr(account, "user", None)
-                user.account = account
-
-        if not account:
-            logger.info("account not found")
-            user = self.user_model.objects.filter(id=account_or_user_id).first()
-            if user:
-                logger.info("user exists: %s", user)
-                user.account = None
-
-        if not user:
-            return None
-        logger.info("authenticate() completed")
-        return (
-            user,
-            header.get("api_key")
-            or (header.get("authorization").split(" ") or ["fake_token"])[-1],
-        )
-
-    def get_header(self, request):
-        """
-        Extracts the header containing the JSON web token from the given
-        request.
-        """
-        account_id = str(request.META.get("HTTP_ACCOUNT", ""))
-        api_key = str(request.META.get("HTTP_API_KEY", ""))
-        authorization = str(request.META.get("HTTP_AUTHORIZATION", ""))
-
-        return dict(account_id=account_id, api_key=api_key, authorization=authorization)
