@@ -1,10 +1,16 @@
 from logging import getLogger
 from typing import Any, Literal
 
-import requests
+from requests import Session
+from requests.adapters import HTTPAdapter, Retry
 
 from .utils import get_service_request_headers
 
+requests = Session()
+retries = Retry(total=2, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
+requests.mount("http://", HTTPAdapter(max_retries=retries))
+
+
 logger = getLogger(__file__)
 
 
