@@ -14,7 +14,7 @@ from .logger import logger
 from .translator import Translator
 from .utils import get_frontend_base_url
 
-DEFAULT_FROM_EMAIL = os.getenv("DEFAULT_FROM_EMAIL", "support@learngual.com")
+DEFAULT_FROM_EMAIL = os.getenv("DEFAULT_FROM_EMAIL", "no-reply@learngual.com")
 
 
 class TransformTemplateContext:
@@ -248,9 +248,20 @@ class TransformTemplateContext:
                 [f'{k}="{v}"' for k, v in attributes["attr"].items()]
             )
             if variable in mapping:
-                html_mapping[variable] = (
+                data_str = (
                     f"<{tag} {attributes_str}>" + "{" + variable + "}" + f"</{tag}>"
-                ).format_map({variable: mapping[variable]})
+                )
+                try:
+                    html_mapping[variable] = data_str.format_map(
+                        {variable: mapping[variable]}
+                    )
+                except KeyError:
+                    raise KeyError(
+                        f"{data_str = }"
+                        f"{variable = }"
+                        "{variable: mapping[variable]} = "
+                        + str({variable: mapping[variable]})
+                    )
         return self.replace_placeholders(
             input_string=plain_text, replacements=html_mapping
         )
@@ -281,43 +292,95 @@ class TransformTemplateContext:
         self,
         emails: list[str],
         context: dict[str, Any],
-        subject,
-        template,
-        fail_silently=...,
-    ):
-        from_email, to = (
-            DEFAULT_FROM_EMAIL,
-            emails,
-        )
-        if isinstance(to, str):
-            to = [to]
-        html_content = render_to_string(template, context)  # render with dynamic value
-        text_content = strip_tags(
-            html_content
-        )  # Strip the html tag. So people can see the pure text at least.
-
-        # create the email, and attach the HTML version as well.
-        msg = EmailMultiAlternatives(subject, text_content, from_email, to)
-        msg.attach_alternative(html_content, "text/html")
-        msg.send(fail_silently=fail_silently)
-
-    def send_email(self, emails: list[str] | str, context={}) -> None:
-        """Endpoint to send email
+        subject: str,
+        template: str,
+        fail_silently: bool = True,
+        from_email: str = DEFAULT_FROM_EMAIL,
+    ) -> None:
+        """
+        Internal method to send an email using Django's EmailMultiAlternatives.
 
         Args:
-            emails (list[str] | str): _description_
-            context (dict, optional): _description_. Defaults to {}.
+            emails (list[str]): List of recipient email addresses.
+            context (dict[str, Any]): Context data for rendering the template.
+            subject (str): Email subject.
+            template (str): Path to the email template.
+            fail_silently (bool, optional): Whether to suppress send errors. Defaults to True.
+            from_email (str, optional): Sender email address. Defaults to DEFAULT_FROM_EMAIL.
+
+        Returns:
+            None
+        """
+        logger.info(
+            f"Rendering email template '{template}' for recipients: {emails} with subject: '{subject}'"
+        )
+        to: list[str] = emails if isinstance(emails, list) else [emails]
+        try:
+            html_content: str = render_to_string(template, context)
+            text_content: str = strip_tags(html_content)
+            msg = EmailMultiAlternatives(subject, text_content, from_email, to)
+            msg.attach_alternative(html_content, "text/html")
+            msg.send(fail_silently=fail_silently)
+            logger.info(
+                f"Email sent to {to} from {from_email} with subject '{subject}'"
+            )
+        except OSError as e:
+            logger.error(f"OS error while sending email to {to}: {e}", exc_info=True)
+            if not fail_silently:
+                raise
+        except ValueError as e:
+            logger.error(f"Value error while sending email to {to}: {e}", exc_info=True)
+            if not fail_silently:
+                raise
+        except TypeError as e:
+            logger.error(f"Type error while sending email to {to}: {e}", exc_info=True)
+            if not fail_silently:
+                raise
+        except Exception as e:
+            # Log unexpected exceptions, but do not raise unless fail_silently is False
+            logger.error(
+                f"Unexpected error while sending email to {to}: {e}", exc_info=True
+            )
+            if not fail_silently:
+                raise
+
+    def send_email(
+        self,
+        emails: list[str] | str,
+        context: dict[str, Any] = {},
+        from_email: str = DEFAULT_FROM_EMAIL,
+    ) -> None:
+        """
+        Send an email using the provided template and context.
+
+        Args:
+            emails (list[str] | str): Recipient email address(es).
+            context (dict, optional): Additional context to merge with the template context. Defaults to {}.
+            from_email (str, optional): Sender email address. Defaults to DEFAULT_FROM_EMAIL.
+
+        Returns:
+            None
         """
         if isinstance(emails, str):
             emails = [emails]
-        context = context or self.get_context()
-        self._send_email(
-            emails=emails,
-            context=context,
-            fail_silently=True,
-            template=self.template,
-            subject=self.subject,
+        logger.info(
+            f"Preparing to send email to: {emails} using template: {self.template} from: {from_email}"
         )
+        try:
+            merged_context = self.get_context() | context
+            self._send_email(
+                from_email=from_email,
+                emails=emails,
+                context=merged_context,
+                fail_silently=True,
+                template=self.template,
+                subject=self.subject,
+            )
+            logger.info(f"Email sent successfully to: {emails}")
+        except (ValueError, KeyError, TypeError, OSError) as e:
+            logger.error(
+                f"Failed to send email to: {emails}. Error: {e}", exc_info=True
+            )
 
 
 class TransformContext:
