@@ -295,7 +295,7 @@ class Translator:
                 )
             # else:
             #     self.add_language_column(target_language.upper())
-            logger.info(f"Language: {target_language} does not exists in sheet")
+            logger.debug(f"Language: {target_language} does not exists in sheet")
             return text
         stripped_text = text.strip()
         translated_dict = self.translations.get(stripped_text)
