@@ -1,4 +1,5 @@
 import importlib
+import json
 import logging
 import mimetypes
 import os
@@ -7,26 +8,39 @@ import subprocess
 import tempfile
 import urllib.parse
 from collections import OrderedDict
+from collections.abc import Callable
 from contextlib import contextmanager
-from datetime import timedelta
+from datetime import datetime, timedelta
 from logging import getLogger
 from pathlib import Path
 from typing import Any, Literal, TypedDict
+from unittest.mock import patch
 from urllib.parse import urlparse
 
 import jwt
 import pytz
-import requests
 from django.conf import settings
 from django.core.cache import cache
 from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile
+from django.db import models
+from django.db.models import Q, QuerySet
+from django.http import HttpRequest, QueryDict
 from django.utils import timezone, translation
+from django_filters.rest_framework import DjangoFilterBackend
 from faker import Faker
+from requests import Session, exceptions
+from requests.adapters import HTTPAdapter, Retry
 from rest_framework.exceptions import ValidationError
+from rest_framework.filters import OrderingFilter, SearchFilter
 
 from .enums import LanguageCodeType
 from .translator import Translator
 
+requests = Session()
+retries = Retry(total=2, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
+requests.mount("http://", HTTPAdapter(max_retries=retries))
+
+
 faker = Faker()
 
 logger = getLogger(__file__)
@@ -42,6 +56,7 @@ LOCAL_URLS = {
     "payment": "http://django-pay:8000",
     "notify": "http://django-notify:8000",
     "media": "http://django-media:8000",
+    "node": "http://node_service:8000",
 }
 
 
@@ -459,7 +474,7 @@ class PermissionManager:
             params (str, optional): e.g name=aka&age=201. Defaults to "".
 
         Raises:
-            requests.exceptions.RequestException: _description_
+            exceptions.RequestException: _description_
 
         Returns:
             _type_: _description_
@@ -480,22 +495,22 @@ class PermissionManager:
         if not res.ok:
             try:
                 response_detail = res.json().get("detail")
-            except requests.JSONDecodeError:
+            except exceptions.JSONDecodeError:
                 response_detail = None
 
             if response_detail == "Invalid service key":
                 logger.error(f"invalid service key, headers={headers}")
-                raise requests.exceptions.RequestException("invalid service key")
+                raise exceptions.RequestException("invalid service key")
             elif response_detail == "account does not exist":
                 logger.error(
                     f"Account does not exist, {response_detail =}, {service =}"
                 )
-                raise requests.exceptions.RequestException(
+                raise exceptions.RequestException(
                     f"Account does not exist in {service} service."
                 )
             else:
                 logger.error("permission service is down %s", res.content)
-                raise requests.exceptions.RequestException(
+                raise exceptions.RequestException(
                     dict(error="permission service is down", response=res.content)
                 )
         return res.json()
@@ -554,26 +569,33 @@ class PermissionManager:
             if not res.ok:
                 try:
                     response_detail = res.json().get("detail")
-                except requests.JSONDecodeError:
+                except exceptions.JSONDecodeError:
                     response_detail = None
 
                 if response_detail == "Invalid service key":
                     logger.error(f"invalid service key, headers={headers}")
-                    raise requests.exceptions.RequestException("invalid service key")
+                    raise exceptions.RequestException("invalid service key")
                 elif response_detail == "account does not exist":
                     logger.error(
                         f"Account does not exist, {response_detail =}, {service =}"
                     )
-                    raise requests.exceptions.RequestException(
+                    raise exceptions.RequestException(
                         f"Account does not exist in {service} service."
                     )
                 else:
                     logger.error("permission service is down %s", res.content)
-                    raise requests.exceptions.RequestException(
+                    raise exceptions.RequestException(
                         dict(error="permission service is down", response=res.content)
                     )
-            data = res.json()
-            cache.set(url_path, data, timeout=cache_timeout)
+            data = {}
+            try:
+                data = res.json()
+                cache.set(url_path, data, timeout=cache_timeout)
+            except exceptions.JSONDecodeError:
+                logger.info(f"Permission service unavailable: {res.content}")
+                raise exceptions.RequestException(
+                    dict(error="permission service is down", response=res.content)
+                )
 
         if dot_path:
             return get_nested_value(data, dot_path)
@@ -581,6 +603,54 @@ class PermissionManager:
 
 
 class TestHelper:
+    permission_request_path = "iam_service.learngual.utils.requests.get"
+    developer_permission = {
+        "metadata": {
+            "is_free": {"value": False},
+            "subscription_expired": {"value": False, "state": "ACTIVE"},
+            "subscription_expired_at": {
+                "value": (timezone.now() + timezone.timedelta(days=30)).isoformat()
+            },
+            "request_count_request_limit_audio_seconds": {
+                "price_per_api_call": 1,
+                "request_limit_count": 23,
+                "excess_price_per_api": 2,
+                "audio_duration": 500,
+            },
+            "IELTS_TOEFL_standard_grading": {"value": True},
+            "ai_speech_analysis": {"value": True},
+            "speech_analysis": {"value": True},
+            "multiple_language_translation": {"Value": "True"},
+            "access_to_learngual_features": {"value": True},
+            "tech_support": {"Value": "True"},
+            "grace_period_ended": {"value": False},
+            "grammar_character_limit": {"value": 4000},
+            "relevance_character_limit": {"value": 4000},
+        }
+    }
+
+    developer_permission_grace_period = {
+        "metadata": {
+            "is_free": {"value": False},
+            "subscription_expired": {"value": True, "state": "GRACE PERIOD"},
+            "subscription_expired_at": {
+                "value": (timezone.now() - timezone.timedelta(days=1)).isoformat()
+            },
+            "request_count_request_limit_audio_seconds": {
+                "price_per_api_call": 1,
+                "request_limit_count": 23,
+                "excess_price_per_api": 2,
+            },
+            "IELTS_TOEFL_standard_grading": {"value": True},
+            "ai_speech_analysis": {"value": True},
+            "speech_analysis": {"value": True},
+            "multiple_language_translation": {"Value": "True"},
+            "access_to_learngual_features": {"value": True},
+            "tech_support": {"Value": "True"},
+            "grace_period_ended": {"value": False},
+        }
+    }
+
     def generate_timedelta(
         self,
         when: Literal["before", "after"],
@@ -660,6 +730,19 @@ class TestHelper:
         query_string = urllib.parse.urlencode(params)
         return f"{url}?{query_string}"
 
+    @contextmanager
+    def mock_permission_context(self, permissions=None):
+        if permissions is None:
+            permissions = {}
+        with patch(self.permission_request_path) as mock_get:
+            mock_get.return_value.status_code = 200
+            mock_get.return_value.ok = True
+            mock_get.return_value.json.return_value = {
+                **self.developer_permission,
+                **permissions,
+            }
+            yield mock_get
+
 
 def get_timezone_from_country(country_code) -> str:
     try:
@@ -677,7 +760,7 @@ def get_base_url(
     service: Literal["iam", "payment", "learn", "notify", "media"] = "iam"
 ):
     url: str = settings.LEARNGUAL_AUTH_RETRIEVE_URL
-    if not url.startswith("https://"):
+    if "learngual.com" not in url:
         url = LOCAL_URLS.get(service) or LOCAL_URLS.get("iam")
 
     parsed_url = urlparse(url)
@@ -904,7 +987,7 @@ def handle_uploaded_file(
         )
 
     # Now call get_media_file_info with the file path
-
+    file.seek(0)
     return media_info
 
 
@@ -937,7 +1020,246 @@ def get_frontend_base_url(
             return f"https://559staging.learngual.com/{path.lstrip('/')}"
         case ("main", "main"):
             return f"https://learngual.com/{path.lstrip('/')}"
-        case ("local", "main"):
-            return f"{base_url}/{path.lstrip('/')}"
+        case ("main", "demo"):
+            return f"https://demo.learngual.com/{path.lstrip('/')}"
+        case ("local", "demo"):
+            return f"https://demo.learngual.com/{path.lstrip('/')}"
         case _:
-            return f"https://learngual.com/{path.lstrip('/')}"
+            return f"https://559staging.learngual.com/{path.lstrip('/')}"
+
+
+class FilterAndSearchManager:
+    def __init__(
+        self,
+        *,
+        request: HttpRequest,
+        filterset_fields: list[str] | dict = [],
+        filterset_keys: dict[str, Callable[[Any], Any | None]] = {},
+        search_fields: list[str] = [],
+        ordering_fields: list[str] = [],
+        ordering: list[str] = [],
+        filter_map: dict[str, str | list[str]] = dict(),
+    ) -> None:
+        """filter, ordering and search management class
+
+        Args:
+            request (HttpRequest): django request object
+            filterset_fields (list[str], optional): a list of fields in the queryset
+            to used and perform filters. Defaults to [].
+            filterset_keys (dict[str,Callable[[Any],Any|None]], optional): \
+                a dict of field_name:converter queryset in the queryset. Defaults to {
+                "rating":int,
+                "price":float
+            }.
+            search_fields (list[str], optional): fields to used for search. Defaults to [].
+            ordering_fields (list[str], optional): fileds to use for default ordering
+            when ordering is not specified. Defaults to [].
+            ordering (list[str], optional): fields allowed to be used for ordering when
+            specifying ordering in the request object. Defaults to [].
+            filter_map (dict[str,str | list[str]], optional): a dictionary where the key is
+              the query params key in the request and the value is the key that will be
+              used for the filter. Defaults to dict().
+
+        Examples:
+            filter_map: {
+                "courses":[
+                    "student_teams__courses__id",
+                    "instructor_teams__courses__id",
+                ],
+                "owner":"courses__owner__id",
+                "rating[]":"courses__rating",
+            }
+        """
+
+        self.filterset_fields = filterset_fields
+        self.search_fields = search_fields
+        self.ordering_fields = ordering_fields
+        self.ordering = ordering
+        self.request = request
+        self.filter_map = filter_map
+        self.filterset_keys = filterset_keys
+
+    filter_backends = [DjangoFilterBackend, SearchFilter, OrderingFilter]
+
+    def subpress_error(
+        self, function: Callable[[Any], Any | None], value: Any
+    ) -> Any | None:
+        try:
+            return function(value)
+        except (ValueError, TypeError):
+            # Ignore errors raised during conversion and continue to the next field.
+            pass
+
+    def build_filter_params(
+        self,
+        request: HttpRequest,
+        field_callable_dict: dict[
+            str, Callable[[Any], Any] | list[Callable[[Any], Any]]
+        ],
+    ) -> dict[str, Any]:
+        """
+        Build filter parameters for Django queryset filter function from request data.
+
+        Parameters:
+            request (django.http.HttpRequest): The Django request object containing the data.
+            field_callable_dict (dict[str, Callable[[Any], Any]|list[Callable[[Any], Any]]]):\
+                 A dictionary with field names as keys and callables as values.
+
+        Returns:
+            dict: A dictionary of filter parameters to be used with Django queryset .filter() function.
+        """
+        filter_params = {}
+
+        if not isinstance(request.GET, QueryDict):
+            return filter_params
+
+        for field, field_callable in field_callable_dict.items():
+            if isinstance(field_callable, list) and len(field_callable) > 0:
+                # If the field_callable is specified as a list, apply each converter to the list of values.
+                values = request.GET.getlist(field)
+                converted_values = [
+                    self.subpress_error(callable_item, val)
+                    for val, callable_item in zip(values, field_callable)
+                ]
+                # Filter out None and empty values from the converted list.
+                converted_values = [val for val in converted_values if val is not None]
+                if converted_values:
+                    filter_params[field] = converted_values
+            else:
+                # If the field_callable is not a list, apply the converter to the single value.
+                value = request.GET.get(field)
+                if value is not None:
+                    try:
+                        converted_value = field_callable(value)
+                        if converted_value is not None:
+                            filter_params[field] = converted_value
+                    except (ValueError, TypeError):
+                        # Ignore errors raised during conversion and continue to the next field.
+                        pass
+
+        return filter_params
+
+    def filter_queryset_with_filterset_keys(self, queryset: QuerySet) -> QuerySet:
+        """method take in a queryset and filter it based on the mapping that was provided
+        with filterset_keys
+
+        filterset_keys:
+        {
+            "name":str,
+            "age":int, #this retrieve it as a single value
+            "ids":[str] #this try retriving the query params as a list
+        }
+
+        Args:
+            queryset (QuerySet):
+
+        Returns:
+            QuerySet:
+        """
+        if not self.filterset_keys:
+            return queryset
+
+        kwargs = self.build_filter_params(self.request, self.filterset_keys)
+        if kwargs:
+            queryset = queryset.filter(**kwargs)
+
+        return queryset
+
+    def filter_queryset_with_dict_maping(self, queryset: QuerySet) -> QuerySet:
+        """method take in a queryset and filter it based on the mapping that was provided
+
+        Args:
+            queryset (QuerySet):
+
+        Returns:
+            QuerySet:
+        """
+        if not self.filter_map:
+            return queryset
+
+        query_params: QueryDict = self.request.GET  # get query dictionary
+        query_kwargs = dict()
+        for key, value in self.filter_map.items():  # loop through the filter map
+            if key in query_params.dict():  # if the filter map key exists
+                if type(value) == str:  # check if it is string
+                    data_list = query_params.getlist(key)
+                    data_single = query_params.get(key)
+                    if data_list:
+                        query_kwargs[f"{value}__in"] = data_list
+                    elif data_single:
+                        query_kwargs[value] = data_single
+                elif type(value) == list:
+                    q = Q()
+                    data_list = query_params.getlist(key)
+                    data_single = query_params.get(key)
+
+                    if data_list:
+                        for sub_value in value:
+                            q |= Q(**{f"{sub_value}__in": data_list})
+                        queryset = queryset.filter(q)
+                    elif data_single:
+                        query_kwargs[value] = query_params.get(key)
+                        for sub_value in value:
+                            q |= Q(**{sub_value: data_single})
+                        queryset = queryset.filter(q)
+        if query_kwargs:
+            queryset = queryset.filter(**query_kwargs)
+
+        return queryset
+
+    def filter_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
+        """
+        Given a queryset, filter it with whichever filter backend is in use.
+
+        You are unlikely to want to override this method, although you may need
+        to call it either from a list view, or from a custom `get_object`
+        method if you want to apply the configured filtering backend to the
+        default queryset.
+        """
+        for Backend in list(self.filter_backends):
+            backend: DjangoFilterBackend = Backend()
+            queryset = backend.filter_queryset(self.request, queryset, self)
+        queryset = self.filter_queryset_with_dict_maping(queryset)
+        queryset = self.filter_queryset_with_filterset_keys(queryset)
+        return queryset
+
+
+class DecodedJWTToken(TypedDict, total=False):
+    exp: str
+    iat: str
+    nbf: str
+    orig_iat: str
+    iss: str
+    sub: str
+    aud: str
+    jti: str
+    typ: str
+    # Allow additional claims
+    # TypedDict allows extra keys by default when total=False
+
+
+def decode_jwt_access_token(token) -> DecodedJWTToken:
+    # Decode the JWT access token
+    decoded_token = jwt.decode(token, options={"verify_signature": False})
+
+    # Convert timestamp values to timedelta objects
+    if "exp" in decoded_token:
+        decoded_token["exp"] = datetime.fromtimestamp(decoded_token["exp"]).isoformat()
+
+    if "iat" in decoded_token:
+        decoded_token["iat"] = datetime.fromtimestamp(decoded_token["iat"]).isoformat()
+    if "nbf" in decoded_token:
+        decoded_token["nbf"] = datetime.fromtimestamp(decoded_token["nbf"]).isoformat()
+    if "orig_iat" in decoded_token:
+        decoded_token["orig_iat"] = datetime.fromtimestamp(
+            decoded_token["orig_iat"]
+        ).isoformat()
+
+    return decoded_token
+
+
+class JSONFieldEncoder(json.JSONEncoder):
+    def default(self, obj):
+        if isinstance(obj, timedelta):
+            return obj.total_seconds()
+        return super().default(obj)
